//*********************************************************//
// Name:      EtherCard + DS18B20 с локациями в EEPROM     //
// Author:    Ilya Danilov www.mk90.ru/blog                //
// Date:      07 May 2013, LAST UPDATE: 18 May 2013        //
// Version:   0.0.1                                        //
// Schematic: Ардуино-совместимая плата на ATmega328P +    //
//            EtherCard (SS=D10) + DS18B20 (D9)            //
// Note:      При обращении через браузер происходит поиск //
//            датчиков DS12B20, считывается и выводится    //
//            текущее значение температуры. В качестве     //
//            идентификатора локации выводится название,   //
//            сохраненное в EEPROM. Название можно         //
//            редактировать через web         См. также    //
//            mk90.blogspot.com/DS18B20                    //    
// Based on EtherCard project: github.com/jcw/ethercard    //
//*********************************************************//
#include <EEPROM.h>
#include <OneWire.h>
#include <EtherCard.h>

// настройки Ethernet

#define BUF_SIZE 512

byte mac[] = { 0x00, 0x04, 0xA3, 0x21, 0xCA, 0x38 }; // MAC-адрес

byte fixed = false; // =false: пробовать получить адрес по DHCP, 
                   //         в случае неудачи использовать статический; 
                   // =true:  сразу использовать статический

uint8_t ip[] = { 169, 254, 8, 200 };     // Статический IP-адрес
uint8_t subnet[] = { 255, 255, 0, 0 };   // Маска подсети
uint8_t gateway[] = { 192, 168, 1, 20 }; // Адрес шлюза 
uint8_t dns[] = { 192, 168, 1, 20 };     // Адрес DNS-сервера (необязателен)

byte Ethernet::buffer[BUF_SIZE];
static BufferFiller bfill;  // used as cursor while filling the buffer

// настройки OneWire

#define DS18B20PIN  9
OneWire ds(DS18B20PIN);

// карта хранения в EEPROM:
// [ 0 ] Число заполненных записей
// [ 1..25 ] Запись №0
//   [ 1..8  ] Адрес 0
//   [ 9..25 ] Имя 0
// [ 26..51 ] Запись №1
// ...

// константы для EEPROM

#define EEFIRSTENTRY  1
#define EEADDRLEN  8
#define EENAMELEN  17

// макросы для доступа в EEPROM

#define LOCADDR(n) (EEFIRSTENTRY+((n)*(EEADDRLEN+EENAMELEN)))
#define LOCNAME(n) (EEFIRSTENTRY+EEADDRLEN+((n)*(EEADDRLEN+EENAMELEN)))

byte searchLocationByAddr(byte *addr) {
  byte total = EEPROM.read(0);
  for (byte i=0;i<total;i++) {
    byte equial = 0;
    for (byte k=0;k<EEADDRLEN;k++) {
      if (addr[k] == EEPROM.read(LOCADDR(i)+k)) equial++;
    }
    if (equial == EEADDRLEN) return i;
  }
  
  // записи не найдено, создаем еще одну - в конце
  for (byte k=0;k<EEADDRLEN;k++) {
    EEPROM.write(LOCADDR(total)+k,addr[k]);
  }
  
  EEPROM.write(LOCNAME(total),'?');
  EEPROM.write(LOCNAME(total)+1,0);
  
  EEPROM.write(0,total+1);
  return total;
}

// команды DS18B20

void setConversionTime(byte conf) {
  ds.reset();
  ds.skip(); // skip ROM
  ds.write(0x4E); // write scratchpad
  ds.write(0); // Th
  ds.write(0); // Tl
  ds.write(conf); // configuration
}

void startConversionAll() {
  ds.reset();
  ds.skip(); // skip ROM
  ds.write(0x44,0); // start conversion
  delay(10);
}

void setup(void)
{
    Serial.begin(57600);
    delay(2000);

    /* Проверяем, что контроллер Ethernet доступен для работы */
    Serial.println("Initialising the Ethernet controller");
    if (ether.begin(sizeof Ethernet::buffer, mac, 10) == 0) {
        Serial.println( "Ethernet controller NOT initialised");
        while (true);
    }

    // Пытаемся получить адрес динамически 
    Serial.println("Attempting to get an IP address using DHCP");
    if (ether.dhcpSetup()) {
        ether.printIp("Got an IP address using DHCP: ", ether.myip);
    } else {
        // Если DHCP не доступен, используем статический ip-адрес
        ether.staticSetup(ip, gateway, dns);
        ether.printIp("DHCP FAILED, using fixed address: ", ether.myip);
        fixed = true;
    }
    
    // считываем и анализируем EEPROM
    if ( (EEPROM.read(0)*(EEADDRLEN+EENAMELEN)+EEFIRSTENTRY) > E2END) {
      EEPROM.write(0,0); // инициализируем число записей
    }
}

char okHeader[] PROGMEM =
  "HTTP/1.0 200 OK\r\n"
  "Content-Type: text/html; charset=utf-8\r\n"
  "Pragma: no-cache\r\n"
;

char authorLink[] PROGMEM =
  "</pre><hr>Read about me <a href=\"http://mk90.blogspot.com\" target=\"_blank\">here</a>"
;

static void homePage (BufferFiller& buf) {
  
  buf.emit_p(PSTR("$F\r\n"
    "<title>Arduino web-thermometr</title>"
    "<h2>DS18B20 Network:</h2>"
    "<pre>"), okHeader);
  
  byte counter = 0; 
  byte addr[8]; 
  byte data[12];
  
  setConversionTime(0x7F); // установить 9-битное разрешение 
  startConversionAll(); // запустить конвертацию температуры
  delay(100); // на конвертацию 9-битного значения требуется 93,75 мс
  
  // ищем устройства и выводим результаты
  ds.reset_search();
  delay(250);
  while (ds.search(addr)) {
    if ( OneWire::crc8(addr, 7) != addr[7]) {
      buf.emit_p(PSTR("- address CRC is not valid!\n"));
      continue;
    } else if (addr[0] != 0x28) {
      buf.emit_p(PSTR("- is not a DS18B20 family device!\n"));
      continue;
    }
    
    counter++;
    byte idx = searchLocationByAddr(addr); // возвращает номер локации в eeprom
    buf.emit_p(PSTR("<a href=\"/e?n=$D\">$E</a>: "), 
      idx, LOCNAME(idx));
    
    ds.reset();
    ds.select(addr);
    ds.write(0xBE); // читать scratchpad
    for (byte k=0; k<9; k++) {  // нам потребуется 9 байт
      data[k] = ds.read();
    }     
    if ( OneWire::crc8( data, 8) != data[8]) {
      buf.emit_p(PSTR("- value CRC is not valid!\n"));
      continue;
    }
    buf.emit_p(PSTR(" $D.$D °C\n"), *(int *)data/16, (int) (abs(*(int *)data % 16) * 0.625));
  }
  
  buf.emit_p(PSTR("\nTotal: $D valid devices."), counter);
  buf.emit_p(authorLink);
}

// обрабатываем передачу параметров методом GET

static void editPage (const char* data, BufferFiller& buf) {
  if (data[6] == '?') {
    char buf2[96];
    if (EtherCard::findKeyVal(data+6, buf2, sizeof(buf2)-1, "n")) {
      byte n = atoi(buf2);
      if (!EtherCard::findKeyVal(data+6, buf2, sizeof(buf2)-1, "s")) {
        // генерировать форму
        buf.emit_p(PSTR("$F\r\n"
        "<h3>Edit location name</h3>"
        "<form>"
         "<p>"
         "Change name: <input type=text name=b value='$E' maxlength=8>"
         "<input type=hidden name=n value=$D>"         
         "<input type=submit name=s value=set>"
         "</p>"
        "</form>"), okHeader, LOCNAME(n), n); 
        return;
      } else {
        // cохранить новое название локации
        byte len = EtherCard::findKeyVal(data+6, buf2, sizeof(buf2)-1, "b");
        if (len) { 
          EtherCard::urlDecode(buf2);
          for (byte i=0;i<(strlen(buf2)+1);i++) 
            EEPROM.write(LOCNAME(n)+i,buf2[i]);
        }
      }
    }
  }  
  buf.emit_p(PSTR(
        "HTTP/1.0 302 found\r\n"
        "Location: /\r\n"
        "\r\n"));  
}

void loop(void)
{
  word len = ether.packetReceive();
  word pos = ether.packetLoop(len);
  if (pos) { 
    bfill = ether.tcpOffset();
    char* data = (char *) Ethernet::buffer + pos;
    Serial.println(data); // распечатываем запрос для отладки
    
    if (strncmp("GET / ", data, 6) == 0) 
      homePage(bfill);
    else if (strncmp("GET /e", data, 6) == 0)
      editPage(data, bfill);
    else
      bfill.emit_p(PSTR(
        "HTTP/1.0 401 Unauthorized\r\n"
        "Content-Type: text/html\r\n"
        "\r\n"
        "<h1>401 Unauthorized</h1>"));
        
    // отправить ответ клиенту
    ether.httpServerReply(bfill.position());
  }
}
